/*
 * Author: Radovan Cerveny
 */

#include "MeasurementProcessor.hpp"
#include <sstream>

namespace measurements {

MeasurementProvisionerResults MeasurementProcessor::process ( const MeasurementProvisionerResults & mpr, const MeasurementProcessorFilter & filter ) {
	MeasurementProvisionerResults newmpr;

	for ( const MPRResults & mprr : mpr.results ) {
		MPRResults newmprr;
		newmprr.inputs = mprr.inputs;

		for ( const MPRPipelineResults & mprpr : mprr.pipelineResults ) {
			MPRPipelineResults newmprpr;
			newmprpr.pipelineStatus = mprpr.pipelineStatus;

			for ( const MPRPipelineCommandResults & mprpcr : mprpr.commandResults ) {

				 // filter out commands
				if ( !filter.matches ( MeasurementProcessorFilter::FilterType::COMMAND, mprpcr.command ) ) continue;

				MPRPipelineCommandResults newmprpcr;
				newmprpcr.command = mprpcr.command;

				newmprpcr.measurementResults = processMeasurementResults ( mprpcr.measurementResults, filter );

				newmprpr.commandResults.push_back ( std::move ( newmprpcr ) );
			}

			if ( newmprpr.commandResults.size ( ) )
				newmprr.pipelineResults.push_back ( std::move ( newmprpr ) );
		}

		if ( newmprr.pipelineResults.size ( ) )
			newmpr.results.push_back ( std::move ( newmprr ) );
	}

	return newmpr;
}

MeasurementResults MeasurementProcessor::processMeasurementResults ( const MeasurementResults & mr, const MeasurementProcessorFilter & filter ) {
	MeasurementResults newmr;

	std::function < void ( unsigned, unsigned ) > dfsLambda;

	dfsLambda = [&] ( unsigned idx, unsigned newmrParentIdx ) {
		if ( idx == 0 ) {
			newmr.frames.push_back ( mr.frames[0] );
			newmr.frames[0].sub_idxs.clear ( );

			for ( unsigned subIdx : mr.frames[idx].sub_idxs )
				dfsLambda ( subIdx, 0 );
		} else {
			const MeasurementFrame & curFrame = mr.frames[idx];

			bool filteredOut = false;
			filteredOut |= !filter.matches ( MeasurementProcessorFilter::FilterType::FRAME_TYPE, to_string ( curFrame.type ) );
			filteredOut |= !filter.matches ( MeasurementProcessorFilter::FilterType::FRAME_NAME, to_string ( curFrame.name ) );

			if ( !filteredOut ) {
				unsigned newmrIdx = newmr.frames.size ( );
				newmr.frames.push_back ( mr.frames[idx] );
				newmr.frames[newmrIdx].sub_idxs.clear ( );
				newmr.frames[newmrParentIdx].sub_idxs.push_back ( newmrIdx );

				for ( unsigned subIdx : mr.frames[idx].sub_idxs )
					dfsLambda ( subIdx, newmrIdx );
			} else {
				for ( unsigned subIdx : mr.frames[idx].sub_idxs )
					dfsLambda ( subIdx, newmrParentIdx );
			}
		}
	};

	dfsLambda ( 0, 0 );

	return newmr;
}

std::string MeasurementProcessor::output ( const MeasurementProvisionerResults & mpr, const MeasurementProcessorOutput & output ) {
	std::stringstream ss;

	MeasurementProcessor::output ( ss, mpr, output );

	return ss.str ( );
}

void MeasurementProcessor::output ( std::ostream & os, const MeasurementProvisionerResults & mpr, const MeasurementProcessorOutput & output ) {
	switch ( output.outputFormat ) {
	case MeasurementProcessorOutput::OutputFormat::XML:
		MeasurementProcessorOutput::outputXml ( os, mpr, output );
		break;

	case MeasurementProcessorOutput::OutputFormat::STATS:
		MeasurementProcessorOutput::outputStats ( os, mpr, output );
		break;

	case MeasurementProcessorOutput::OutputFormat::CSV:
		MeasurementProcessorOutput::outputCsv ( os, mpr, output );
		break;

	case MeasurementProcessorOutput::OutputFormat::HTML:
		MeasurementProcessorOutput::outputHtml ( os, mpr, output );
		break;
	}
}

}
