/*
 * Author: Radovan Cerveny
 */

#include "MeasurementProcessorOutput.hpp"
#include <set>
#include <iostream>

namespace measurements {

void MeasurementProcessorOutput::outputXml ( std::ostream & os, const MeasurementProvisionerResults & mpr, const MeasurementProcessorOutput & ) {
	os << mpr;
}

void MeasurementProcessorOutput::outputStats ( std::ostream & os, const MeasurementProvisionerResults & mpr, const MeasurementProcessorOutput & ) {

	std::vector < int > errorExitCodes;
	std::set < std::string > measuredPipelines;

	for ( const MPRResults & mprr : mpr.results )
		for ( const MPRPipelineResults & mprpr : mprr.pipelineResults ) {
			if ( mprpr.pipelineStatus.exitCode != 0 )
				errorExitCodes.push_back ( mprpr.pipelineStatus.exitCode );

			std::string pipeline;

			for ( auto it = mprpr.commandResults.cbegin ( ); it != mprpr.commandResults.cend ( ); ++it ) {
				if ( it != mprpr.commandResults.cbegin ( ) )
					pipeline += " | ";

				pipeline += it->command;
			}

			measuredPipelines.insert ( pipeline );
		}

	os << "Total measurements: " << mpr.results.size ( ) << std::endl;
	os << "Errors: ";

	if ( errorExitCodes.size ( ) == 0 )
		os << "none" << std::endl;
	else
		os << errorExitCodes << std::endl;

	os << "Measured pipelines:" << std::endl;

	for ( const std::string & pipeline : measuredPipelines )
		os << "  " << pipeline << std::endl;
}

void MeasurementProcessorOutput::outputCsv ( std::ostream & os, const MeasurementProvisionerResults & mpr, const MeasurementProcessorOutput & mpo ) {

	struct Row {
		std::string					input;
		std::vector < std::string > data;
	};

	struct Table {
		std::string					description;
		std::vector < std::string > header;
		std::vector < Row >			rows;
	};

	 // this starts to get ugly
	std::vector < std::string > masterHeader {
		"input"
	};

	for ( const MPRPipelineResults & mprpr : mpr.results[0].pipelineResults )
		for ( const MPRPipelineCommandResults & mprpcr : mprpr.commandResults )
			masterHeader.push_back ( mprpcr.command );

	 // this is definitely ugly
	const auto & masterFrames = mpr.results[0].pipelineResults[0].commandResults[0].measurementResults.frames;

	std::vector < Table > tables ( masterFrames.size ( ) - 1 );

	 // start from 1 = skip ROOT
	for ( size_t frameIdx = 1; frameIdx < masterFrames.size ( ); ++frameIdx ) {
		Table & table = tables[frameIdx - 1];

		table.description = to_string ( masterFrames[frameIdx].name ) + " " + to_string ( masterFrames[frameIdx].type );
		table.header = masterHeader;

		for ( const MPRResults & mprr : mpr.results ) {
			Row row;

			for ( const auto & input : mprr.inputs )
				row.input += input.second + " ";

			for ( const MPRPipelineResults & mprpr : mprr.pipelineResults )
				for ( const MPRPipelineCommandResults & mprpcr : mprpr.commandResults )
					row.data.push_back ( MeasurementProcessorOutput::getData ( mprpcr.measurementResults.frames[frameIdx], mpo ) );

			table.rows.push_back ( std::move ( row ) );
		}
	}

	for ( const Table & table : tables ) {
		os << "# " << table.description << std::endl;

		for ( auto it = table.header.begin ( ); it != table.header.end ( ); ++it ) {
			if ( it != table.header.begin ( ) ) os << ",";

			os << * it;
		}

		os << std::endl;

		for ( const Row & row : table.rows ) {
			os << row.input;

			for ( const std::string & datum : row.data )
				os << "," << datum;

			os << std::endl;
		}
	}
}

void MeasurementProcessorOutput::outputHtml ( std::ostream &, const MeasurementProvisionerResults &, const MeasurementProcessorOutput & ) {
}

std::string MeasurementProcessorOutput::getData ( const MeasurementFrame & frame, const MeasurementProcessorOutput & mpo ) {

	switch ( mpo.outputEngine ) {
	case OutputEngine::TIME:
		return std::to_string ( frame.time.duration.count ( ) );

		break;

	case OutputEngine::MEMORY:
		return std::to_string ( frame.memory.high_watermark - frame.memory.start_heap_usage );

		break;

	case OutputEngine::COUNTER:
		return "";

		break;
	}

	return "";
}

}
