/*
 * Author: Radovan Cerveny
 */

#include "sax/FromXMLParserHelper.h"
#include "MeasurementProvisionerConfigurationXml.hpp"
#include "exception/CommonException.h"

using namespace std;
using namespace sax;

namespace measurements {

MeasurementProvisionerConfiguration MeasurementProvisionerConfigurationXml::parse ( deque < Token > & tokens ) {
	MeasurementProvisionerConfiguration configuration;

	deque < Token >::iterator iter = tokens.begin ( );

	if ( iter == tokens.end ( ) ) throw::exception::CommonException ( "Empty tokens list" );

	parseMeasurementProvisioner ( configuration, iter );

	if ( iter != tokens.end ( ) ) throw::exception::CommonException ( "Unexpeted tokens at the end of the xml" );

	return configuration;
}

void MeasurementProvisionerConfigurationXml::parseMeasurementProvisioner ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, MEASUREMENT_PROVISIONER_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == ENVIRONMENT_TAG )
			parseEnvironment ( mpc, iter );
		else if ( data == INPUT_DATA_TAG )
			parseInputData ( mpc, iter );
		else if ( data == PIPELINES_TAG )
			parsePipelines ( mpc, iter );
		else
			break;
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, MEASUREMENT_PROVISIONER_TAG );
}

void MeasurementProvisionerConfigurationXml::parseEnvironment ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, ENVIRONMENT_TAG );

	mpc.environment.workingDirectory = "";
	mpc.environment.pipelineIterations = 1;

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == WORKING_DIRECTORY_TAG ) {
			FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, WORKING_DIRECTORY_TAG );
			mpc.environment.workingDirectory = FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER );
			FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, WORKING_DIRECTORY_TAG );
		} else if ( data == PIPELINE_ITERATIONS_TAG ) {
			FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, PIPELINE_ITERATIONS_TAG );
			mpc.environment.pipelineIterations = std::stoi ( FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER ) );
			FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, PIPELINE_ITERATIONS_TAG );
		} else {
			break;
		}
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, ENVIRONMENT_TAG );
}

void MeasurementProvisionerConfigurationXml::parseInputData ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, INPUT_DATA_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == INPUT_DATA_BATCH_TAG )
			parseInputBatch ( mpc, iter );
		else
			break;
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, INPUT_DATA_TAG );
}

void MeasurementProvisionerConfigurationXml::parseInputBatch ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, INPUT_DATA_BATCH_TAG );

	MPCInputBatch mpcib;

	mpcib.forcedDependencyBatch = false;

	if ( FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ATTRIBUTE ) ) {
		string attrName = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ATTRIBUTE );

		if ( attrName == INPUT_DATA_BATCH_ATTR_DEPENDENCY_NAME ) {
			FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ATTRIBUTE, attrName );
			string attrVal = FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER );

			if ( attrVal == "true" )
				mpcib.forcedDependencyBatch = true;
			else if ( attrVal == "false" )
				mpcib.forcedDependencyBatch = false;
			else
				throw::exception::CommonException ( "XML: MPCInputBatch malformed boolean attr: " + attrName );

			FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ATTRIBUTE, attrName );
		}
	}

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		mpcib.batch.push_back ( parseInputDatum ( data, iter ) );
	}

	mpc.inputBatches.push_back ( std::move ( mpcib ) );

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, INPUT_DATA_BATCH_TAG );
}

MPCInputDatum MeasurementProvisionerConfigurationXml::parseInputDatum ( const std::string & tag, std::deque < sax::Token >::iterator & iter ) {
	MPCInputDatum mpcid;

	mpcid.id = -1;
	mpcid.count = 1;

	if ( tag == INPUT_DATA_FILE_TAG )
		mpcid.type = MPCInputDatum::Type::FILE;
	else if ( tag == INPUT_DATA_GENERATOR_TAG )
		mpcid.type = MPCInputDatum::Type::GENERATOR;
	else
		throw::exception::CommonException ( "XML: Unknown MPCInputDatum type" );

	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, tag );

	bool foundId = false;

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ATTRIBUTE ) ) break;

		string attrName = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ATTRIBUTE );
		FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ATTRIBUTE, attrName );
		string attrVal = FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER );
		FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ATTRIBUTE, attrName );

		if ( attrName == INPUT_DATA_ATTR_ID_NAME ) {
			mpcid.id = std::stoi ( attrVal );
			foundId	 = true;
		} else if ( attrName == INPUT_DATA_ATTR_COUNT_NAME ) {
			mpcid.count = std::stoi ( attrVal );
		} else {
			throw::exception::CommonException ( "XML: MPCInputDatum unknown attr: " + attrName );
		}
	}

	if ( !foundId )
		throw::exception::CommonException ( "XML: MPCInputDatum id attr was not specified" );

	mpcid.value = FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER );

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, tag );
	return mpcid;
}

void MeasurementProvisionerConfigurationXml::parsePipelines ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, PIPELINES_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == PIPELINE_TAG ) {
			mpc.pipelines.push_back ( { } );
			parsePipeline ( mpc.pipelines.back ( ), iter );
		} else {
			break;
		}
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, PIPELINES_TAG );
}

void MeasurementProvisionerConfigurationXml::parsePipeline ( MPCPipeline & pipeline, std::deque < sax::Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, PIPELINE_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == COMMAND_TAG )
			pipeline.push_back ( parsePipelineCommand ( iter ) );
		else
			break;
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, PIPELINE_TAG );
}

MPCPipelineCommand MeasurementProvisionerConfigurationXml::parsePipelineCommand ( std::deque < sax::Token >::iterator & iter ) {

	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, COMMAND_TAG );

	MPCPipelineCommand mpcpc;

	mpcpc.measure = true;

	if ( FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ATTRIBUTE ) ) {
		string attrName = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ATTRIBUTE );

		if ( attrName == COMMAND_ATTR_MEASURE_NAME ) {
			FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ATTRIBUTE, attrName );
			string attrVal = FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER );

			if ( attrVal == "true" )
				mpcpc.measure = true;
			else if ( attrVal == "false" )
				mpcpc.measure = false;
			else
				throw::exception::CommonException ( "XML: MPCPipelineCommand malformed boolean attr: " + attrName );

			FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ATTRIBUTE, attrName );
		}
	}

	mpcpc.command = FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER );
	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, COMMAND_TAG );

	return mpcpc;
}

void MeasurementProvisionerConfigurationXml::parseMeasurementConfiguration ( MeasurementProvisionerConfiguration &, std::deque < sax::Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, MEASUREMENT_CONFIGURATION_TAG );

	 // skip everything for now
	while ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::END_ELEMENT ) || FromXMLParserHelper::getTokenData ( iter, Token::TokenType::END_ELEMENT ) != MEASUREMENT_CONFIGURATION_TAG ) ++iter;

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, MEASUREMENT_CONFIGURATION_TAG );
}

string MeasurementProvisionerConfigurationXml::MEASUREMENT_PROVISIONER_TAG = "MeasurementProvisioner";

string MeasurementProvisionerConfigurationXml::ENVIRONMENT_TAG = "Environment";
string MeasurementProvisionerConfigurationXml::WORKING_DIRECTORY_TAG = "WorkingDirectory";
string MeasurementProvisionerConfigurationXml::PIPELINE_ITERATIONS_TAG = "PipelineIterations";

string MeasurementProvisionerConfigurationXml::INPUT_DATA_TAG = "InputData";
string MeasurementProvisionerConfigurationXml::INPUT_DATA_BATCH_TAG = "InputBatch";
string MeasurementProvisionerConfigurationXml::INPUT_DATA_BATCH_ATTR_DEPENDENCY_NAME = "dependency";

string MeasurementProvisionerConfigurationXml::INPUT_DATA_FILE_TAG = "File";
string MeasurementProvisionerConfigurationXml::INPUT_DATA_GENERATOR_TAG = "Generator";

string MeasurementProvisionerConfigurationXml::INPUT_DATA_ATTR_ID_NAME = "id";
string MeasurementProvisionerConfigurationXml::INPUT_DATA_ATTR_COUNT_NAME = "count";

string MeasurementProvisionerConfigurationXml::PIPELINES_TAG = "Pipelines";
string MeasurementProvisionerConfigurationXml::PIPELINE_TAG	 = "Pipeline";
string MeasurementProvisionerConfigurationXml::COMMAND_TAG = "Command";
string MeasurementProvisionerConfigurationXml::COMMAND_ATTR_MEASURE_NAME = "measure";

string MeasurementProvisionerConfigurationXml::MEASUREMENT_CONFIGURATION_TAG = "MeasurementConfiguration";
}
