/*
 * Author: Radovan Cerveny
 */

#include "MeasurementProvisionerResultsXml.hpp"
#include "exception/CommonException.h"
#include "sax/FromXMLParserHelper.h"
#include <measure>

using namespace std;
using namespace sax;

namespace measurements {

deque < Token > MeasurementProvisionerResultsXml::compose ( const MeasurementProvisionerResults & mpr ) {
	deque < Token > tokens;

	tokens.emplace_back ( MEASUREMENT_PROVISIONER_RESULTS_TAG, Token::TokenType::START_ELEMENT );

	for ( const MPRResults & mprr : mpr.results )
		composeMeasurementProvisionerResult ( tokens, mprr );

	tokens.emplace_back ( MEASUREMENT_PROVISIONER_RESULTS_TAG, Token::TokenType::END_ELEMENT );
	return tokens;
}

void MeasurementProvisionerResultsXml::composeMeasurementProvisionerResult ( deque < Token > & tokens, const MPRResults & mprr ) {
	tokens.emplace_back ( MEASUREMENT_PROVISIONER_RESULT_TAG, Token::TokenType::START_ELEMENT );

	composeInputs ( tokens, mprr );
	composePipelines ( tokens, mprr );

	tokens.emplace_back ( MEASUREMENT_PROVISIONER_RESULT_TAG, Token::TokenType::END_ELEMENT );
}

void MeasurementProvisionerResultsXml::composeInputs ( deque < Token > & tokens, const MPRResults & mprr ) {
	tokens.emplace_back ( INPUTS_TAG, Token::TokenType::START_ELEMENT );

	for ( const std::pair < int, std::string > & input : mprr.inputs ) {
		tokens.emplace_back ( INPUT_TAG, Token::TokenType::START_ELEMENT );
		tokens.emplace_back ( INPUT_ATTR_ID_NAME, Token::TokenType::START_ATTRIBUTE );
		tokens.emplace_back ( std::to_string ( input.first ), Token::TokenType::CHARACTER );
		tokens.emplace_back ( INPUT_ATTR_ID_NAME, Token::TokenType::END_ATTRIBUTE );
		tokens.emplace_back ( input.second, Token::TokenType::CHARACTER );
		tokens.emplace_back ( INPUT_TAG, Token::TokenType::END_ELEMENT );
	}

	tokens.emplace_back ( INPUTS_TAG, Token::TokenType::END_ELEMENT );
}

void MeasurementProvisionerResultsXml::composePipelines ( deque < Token > & tokens, const MPRResults & mprr ) {
	tokens.emplace_back ( PIPELINES_TAG, Token::TokenType::START_ELEMENT );

	for ( const MPRPipelineResults & mprpr : mprr.pipelineResults )
		composePipeline ( tokens, mprpr );

	tokens.emplace_back ( PIPELINES_TAG, Token::TokenType::END_ELEMENT );
}

void MeasurementProvisionerResultsXml::composePipeline ( deque < Token > & tokens, const MPRPipelineResults & mprpr ) {
	tokens.emplace_back ( PIPELINE_TAG, Token::TokenType::START_ELEMENT );

	composePipelineStatus ( tokens, mprpr );
	composeCommands ( tokens, mprpr );

	tokens.emplace_back ( PIPELINE_TAG, Token::TokenType::END_ELEMENT );
}

void MeasurementProvisionerResultsXml::composePipelineStatus ( deque < Token > & tokens, const MPRPipelineResults & mprpr ) {

	tokens.emplace_back ( PIPELINE_STATUS_TAG, Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( EXIT_CODE_TAG, Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( std::to_string ( mprpr.pipelineStatus.exitCode ), Token::TokenType::CHARACTER );
	tokens.emplace_back ( EXIT_CODE_TAG, Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( ERROR_ORIGIN_TAG, Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( mprpr.pipelineStatus.errorOrigin, Token::TokenType::CHARACTER );
	tokens.emplace_back ( ERROR_ORIGIN_TAG, Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( ERROR_VALUE_TAG, Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( mprpr.pipelineStatus.errorValue, Token::TokenType::CHARACTER );
	tokens.emplace_back ( ERROR_VALUE_TAG, Token::TokenType::END_ELEMENT );

	tokens.emplace_back ( PIPELINE_STATUS_TAG, Token::TokenType::END_ELEMENT );
}

void MeasurementProvisionerResultsXml::composeCommands ( deque < Token > & tokens, const MPRPipelineResults & mprpr ) {
	tokens.emplace_back ( COMMANDS_TAG, Token::TokenType::START_ELEMENT );

	for ( const MPRPipelineCommandResults & mprpcr : mprpr.commandResults )
		composeCommand ( tokens, mprpcr );

	tokens.emplace_back ( COMMANDS_TAG, Token::TokenType::END_ELEMENT );
}

void MeasurementProvisionerResultsXml::composeCommand ( deque < Token > & tokens, const MPRPipelineCommandResults & mprpcr ) {
	tokens.emplace_back ( COMMAND_TAG, Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( COMMAND_NAME_TAG, Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( mprpcr.command, Token::TokenType::CHARACTER );
	tokens.emplace_back ( COMMAND_NAME_TAG, Token::TokenType::END_ELEMENT );

	measurements::MeasurementResultsXml::compose ( mprpcr.measurementResults, tokens );

	tokens.emplace_back ( COMMAND_TAG, Token::TokenType::END_ELEMENT );
}

/*****************************************************************************************************/

MeasurementProvisionerResults MeasurementProvisionerResultsXml::parse ( deque < Token > & tokens ) {
	MeasurementProvisionerResults mprs;

	deque < Token >::iterator iter = tokens.begin ( );

	if ( iter == tokens.end ( ) ) throw::exception::CommonException ( "Empty tokens list" );

	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, MEASUREMENT_PROVISIONER_RESULTS_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == MEASUREMENT_PROVISIONER_RESULT_TAG )
			mprs.results.push_back ( parseMeasurementProvisionerResult ( iter ) );
		else
			break;
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, MEASUREMENT_PROVISIONER_RESULTS_TAG );

	if ( iter != tokens.end ( ) ) throw::exception::CommonException ( "Unexpeted tokens at the end of the xml" );

	return mprs;
}

MPRResults MeasurementProvisionerResultsXml::parseMeasurementProvisionerResult ( deque < Token >::iterator & iter ) {
	MPRResults mprr;

	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, MEASUREMENT_PROVISIONER_RESULT_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == INPUTS_TAG )
			parseInputs ( mprr, iter );
		else if ( data == PIPELINES_TAG )
			parsePipelines ( mprr, iter );
		else
			break;
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, MEASUREMENT_PROVISIONER_RESULT_TAG );

	return mprr;
}

void MeasurementProvisionerResultsXml::parseInputs ( MPRResults & mprr, deque < Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, INPUTS_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, INPUT_TAG );

		string idAttr = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ATTRIBUTE );

		if ( idAttr != INPUT_ATTR_ID_NAME )
			throw::exception::CommonException ( "MPRResults Input expects only id attr" );

		FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ATTRIBUTE, idAttr );

		int id = stoi ( FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER ) );

		FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ATTRIBUTE, idAttr );

		string inputName = FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER );

		mprr.inputs.emplace_back ( id, std::move ( inputName ) );

		FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, INPUT_TAG );
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, INPUTS_TAG );
}

void MeasurementProvisionerResultsXml::parsePipelines ( MPRResults & mprr, deque < Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, PIPELINES_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == PIPELINE_TAG )
			mprr.pipelineResults.push_back ( parsePipeline ( iter ) );
		else
			break;
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, PIPELINES_TAG );
}

MPRPipelineResults MeasurementProvisionerResultsXml::parsePipeline ( deque < Token >::iterator & iter ) {
	MPRPipelineResults mprpr;
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, PIPELINE_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == PIPELINE_STATUS_TAG )
			parsePipelineStatus ( mprpr, iter );
		else if ( data == COMMANDS_TAG )
			parseCommands ( mprpr, iter );
		else
			break;
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, PIPELINE_TAG );

	return mprpr;
}

void MeasurementProvisionerResultsXml::parsePipelineStatus ( MPRPipelineResults & mprpr, deque < Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, PIPELINE_STATUS_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == EXIT_CODE_TAG ) {
			FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, EXIT_CODE_TAG );
			mprpr.pipelineStatus.exitCode = stoi ( FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER ) );
			FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, EXIT_CODE_TAG );
		} else if ( data == ERROR_ORIGIN_TAG ) {
			FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, ERROR_ORIGIN_TAG );

			if ( FromXMLParserHelper::isTokenType ( iter, Token::TokenType::CHARACTER ) )
				mprpr.pipelineStatus.errorOrigin = FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER );

			FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, ERROR_ORIGIN_TAG );
		} else if ( data == ERROR_VALUE_TAG ) {
			FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, ERROR_VALUE_TAG );

			if ( FromXMLParserHelper::isTokenType ( iter, Token::TokenType::CHARACTER ) )
				mprpr.pipelineStatus.errorValue = FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER );

			FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, ERROR_VALUE_TAG );
		} else {
			break;
		}
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, PIPELINE_STATUS_TAG );
}

void MeasurementProvisionerResultsXml::parseCommands ( MPRPipelineResults & mprpr, deque < Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, COMMANDS_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == COMMAND_TAG )
			mprpr.commandResults.push_back ( parseCommand ( iter ) );
		else
			break;
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, COMMANDS_TAG );
}

MPRPipelineCommandResults MeasurementProvisionerResultsXml::parseCommand ( deque < Token >::iterator & iter ) {
	MPRPipelineCommandResults mprpcr;
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, COMMAND_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == COMMAND_NAME_TAG ) {
			FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, COMMAND_NAME_TAG );
			mprpcr.command = FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER );
			FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, COMMAND_NAME_TAG );
		} else if ( data == MEASUREMENT_RESULTS_TAG ) {
			mprpcr.measurementResults = measurements::MeasurementResultsXml::parse ( iter );
		} else {
			break;
		}
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, COMMAND_TAG );
	return mprpcr;
}

/*****************************************************************************************************/

string MeasurementProvisionerResultsXml::MEASUREMENT_PROVISIONER_RESULTS_TAG = "MeasurementProvisionerResults";

string MeasurementProvisionerResultsXml::MEASUREMENT_PROVISIONER_RESULT_TAG = "MeasurementProvisionerResult";

string MeasurementProvisionerResultsXml::INPUTS_TAG = "Inputs";

string MeasurementProvisionerResultsXml::INPUT_TAG = "Input";
string MeasurementProvisionerResultsXml::INPUT_ATTR_ID_NAME = "id";

string MeasurementProvisionerResultsXml::PIPELINES_TAG = "Pipelines";
string MeasurementProvisionerResultsXml::PIPELINE_TAG  = "Pipeline";

string MeasurementProvisionerResultsXml::PIPELINE_STATUS_TAG = "PipelineStatus";
string MeasurementProvisionerResultsXml::EXIT_CODE_TAG = "ExitCode";
string MeasurementProvisionerResultsXml::ERROR_ORIGIN_TAG = "ErrorOrigin";
string MeasurementProvisionerResultsXml::ERROR_VALUE_TAG  = "ErrorValue";

string MeasurementProvisionerResultsXml::COMMANDS_TAG = "Commands";
string MeasurementProvisionerResultsXml::COMMAND_TAG  = "Command";

string MeasurementProvisionerResultsXml::COMMAND_NAME_TAG = "Name";

string MeasurementProvisionerResultsXml::MEASUREMENT_RESULTS_TAG = "MeasurementResults";
}
